# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import ChartInfoClass
from .....schema_classes import ChartQueryClass
from .....schema_classes import ChartQueryTypeClass
from .....schema_classes import ChartTypeClass
from .....schema_classes import ChartUsageStatisticsClass
from .....schema_classes import ChartUserUsageCountsClass
from .....schema_classes import EditableChartPropertiesClass


ChartInfo = ChartInfoClass
ChartQuery = ChartQueryClass
ChartQueryType = ChartQueryTypeClass
ChartType = ChartTypeClass
ChartUsageStatistics = ChartUsageStatisticsClass
ChartUserUsageCounts = ChartUserUsageCountsClass
EditableChartProperties = EditableChartPropertiesClass

# fmt: on
